/****************************************************************/
/* E-CPU2 v0.2                                                  */
/*   author: Shawn M. Veader                                    */
/*   email: veader@cc.gatech.edu                                */
/****************************************************************/
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <sys/wait.h>
#include <errno.h>
#include <epplet.h>

int					timercount = 0;
int					cpus = 0;
int					cur_cpu = 0;
double             *prev_val = NULL;
int                *load_val = NULL;
unsigned long		prevTotal = 0;
unsigned long		prevOut = 0;
unsigned long		prevIn = 0;
Window              win;
RGB_buf             buf;
Epplet_gadget       da, b_close, b_config, b_help, pop, text, txtpercent;
int                *flame = NULL;
int                *vspread, *hspread, *residual;
unsigned char       rm[255], gm[255], bm[255];
double stream_max   = 2000;

static int          colors[] =
{
   30, 90,  90,
   50, 255, 255,
   255, 255, 255,

   255, 0,  0,
   255, 255, 0,
   255, 255, 255,

   0, 255,  0,
   255, 255, 0,
   255, 255, 255,

   0, 0,  255,
   255, 0, 255,
   255, 255, 255,

   0, 0,  200,
   40, 80, 255,
   100, 200, 255,

   80,  90,  140,
   140, 150, 180,
   255, 230, 200,

   20,  40,  180,
   255, 160, 0,
   255, 255, 100

};

static void save_conf(int d1, int d2, int d3, int d4, int d5, 
				      int d6, int d7, int d8, int d9);
static void load_conf(void);
static void epplet_in(void *data, Window w);
static void epplet_out(void *data, Window w);
static void epplet_timer(void *data);
static void epplet_close(void *data);
static void epplet_config(void *data);
static void epplet_help(void *data);
static int  count_cpus(void);
static void draw_flame(void);
static void flame_col(int r1, int g1, int b1, 
				      int r2, int g2, int b2, 
					  int r3, int g3, int b3);

#define VARIANCE 40
#define VARTREND 16

#define HSPREAD  10
#define VSPREAD  160
#define RESIDUAL 75

#define DIVISIONS 1

#define WIDTH  40
#define HEIGHT 30

#define MAX      255

/* set the flame color */
static void flame_col(int r1, int g1, int b1, 
					  int r2, int g2, int b2, 
					  int r3, int g3, int b3)
{
	int i;
   
	for(i = 0; i < 25; i++) {
		rm[i] = (r1 * i) / 25;
		gm[i] = (g1 * i) / 25;
		bm[i] = (b1 * i) / 25;
	}
	for(i = 0; i < 25; i++) {
		rm[25 + i] = ((r2 * i) + (r1 * (25 - i))) / 25;
		gm[25 + i] = ((g2 * i) + (g1 * (25 - i))) / 25;
		bm[25 + i] = ((b2 * i) + (b1 * (25 - i))) / 25;
	}
	for(i = 0; i < 25; i++) {
		rm[50 + i] = ((r3 * i) + (r2 * (25 - i))) / 25;
		gm[50 + i] = ((g3 * i) + (g2 * (25 - i))) / 25;
		bm[50 + i] = ((b3 * i) + (b2 * (25 - i))) / 25;
 	}

	for(i = 75; i < 255; i++) {
		rm[i] = rm[74];
		gm[i] = gm[74];
		bm[i] = bm[74];
 	}
}

/* draw the flame to display */
static void draw_flame(void)
{
	unsigned char *rgb, *rptr;
	int x, y, *ptr, val1, val2, val3, i, j;
  
	/* initialize the flame if it isn't done already */
	if(!flame) {
		vspread = malloc(40 * sizeof(int));
		hspread = malloc(40 * sizeof(int));
		residual = malloc(40 * sizeof(int));
		flame = malloc(sizeof(int) * WIDTH * HEIGHT);
		memset(flame, 0, sizeof(int) * WIDTH * HEIGHT);
	}

	/* move to the bottom left of the drawing area */
	ptr = flame + ((HEIGHT - 1) * WIDTH);
	/* scan along the bottom row of pixels to start flames */
	for(x = 0; x < WIDTH; x++) {
		/* adjust spreads and residual values to reflect the
		 * load values... */
		vspread[x] = VSPREAD + (load_val[(x * DIVISIONS) / HEIGHT] / 50);
		hspread[x] = HSPREAD + (load_val[(x * DIVISIONS) / WIDTH] / 50);
		residual[x] = RESIDUAL + (load_val[(x * DIVISIONS) / HEIGHT] / 50);
		/* assign a random value to the pixel according to the
		 * load ... gives randomness to flames */
		ptr[x] = (rand() % ((load_val[(x * DIVISIONS) / 40]) + 155));
		/* bounds checking */
		if(ptr[x] > MAX) {
			ptr[x] = 0;
		} else if(ptr[x] < 0) {
			ptr[x] = 0;
		}
	}
   
	/* this divides up the chart into multiple sections
	 * depending on the cpus value */
	for(x = (WIDTH / (DIVISIONS * 2)) - 1 + (0 * (WIDTH / DIVISIONS));
		x <= (WIDTH / (DIVISIONS * 2)) + 1 + (0 * (WIDTH / DIVISIONS));
		x++)
	{
		j = (load_val[cur_cpu] * HEIGHT) / 100;
		ptr = flame + ((HEIGHT - j) * (WIDTH)) + (x);
		/* marches down a column increasing the intensity
		 * as you travel down */
		for(y = 0; y < j; y++) {
			ptr[0] += ((y * 64 * load_val[cur_cpu]) / (j * 100));
			ptr += WIDTH;
		}
	}
		
	/* ----- *
	 * --A-- *
	 * -BCD- *
	 * ----- */
	for(y = HEIGHT - 1; y >= 2; y--) {	
		ptr = flame + (y * 40);
		for(x = 1; x < WIDTH - 1; x++) {
			/* val1 = (C * vspread) / 256 */
			val1 = (ptr[x] * vspread[x]) >> 8;
			/* val2 = (C * hspread) / 256 */
			val2 = (ptr[x] * hspread[x]) >> 8;
			/* val3 = (C * residual) / 256 */
			val3 = (ptr[x] * residual[x]) >> 8;
		
			/* add val2 to B */
			ptr[x - 1] += val2;
			if(ptr[x - 1] > MAX)
				ptr[x - 1] = MAX;
			/* add val1 to A */
			ptr[x - WIDTH] += val1;
			if(ptr[x - WIDTH] > MAX)
				ptr[x - WIDTH] = MAX;
			/* add val2 to D */
			ptr[x + 1] += val2;
			if(ptr[x + 1] > MAX)
				ptr[x + 1] = MAX;
			/* add val3 to C */
			ptr[x] = val3;
		}
	}

	/* blank the top row of pixels */
	for(x = 0; x < WIDTH; x++)
		flame[x] = 0;
	/* half the intensity of the second row */
	for(x = 0; x < WIDTH; x++)
 		flame[WIDTH + x] /= 2;
	/* blank out left most column */
 	for(y = 0; y < HEIGHT; y++)
		flame[y * WIDTH] = 0;
	/* half the intensity on second left most column */
	for(y = 0; y < HEIGHT; y++)
		flame[(y * WIDTH) + 1] /= 2;
	/* blank out right most column */
	for(y = 0; y < HEIGHT; y++)
		flame[(y * WIDTH) + (WIDTH - 1)] = 0;
	/* half the intensity of the second right most column */
	for(y = 0; y < HEIGHT; y++)
		flame[(y * WIDTH) + (WIDTH - 2)] /= 2;


	rgb = Epplet_get_rgb_pointer(buf);
	for(y = 0; y < HEIGHT; y++) {
		ptr = flame + (y * WIDTH) + 1;
		rptr = rgb + (y * WIDTH * 3);
		for(x = 0; x < WIDTH; x++) {
			val1 = ptr[x] & 0xff;
			rptr[0] = rm[val1];
			rptr[1] = gm[val1];
			rptr[2] = bm[val1];
			rptr += 3;
		}
	}
}

/* handles the timer, check /proc/stat for CPU usage */
static void cb_timer(void *data)
{
	static FILE *f;
	int i;
	
	f = fopen("/proc/stat", "r");
	if (f) {
		char s[256];
		if (cpus > 1)
			fgets(s, 255, f);
		for (i = 0; i < cpus; i++) {
			unsigned long user, nice, sys, idle;
			char ss[64];
			double val, val2;
			fgets(s, 255, f);
			sscanf(s, "%*s %lu %lu %lu %lu", &user, &nice, &sys, &idle);
			val = (double) user;
			val2 = val - prev_val[i];
			prev_val[i] = val;
			val2 *= 10;
			if (val2 > 100)
				val2 = 100;
			load_val[i] = val2;

			/* put up the percentage for the correct cpu
			 * assume one cpu percentage per instance */
			if(i == cur_cpu) {
				if(timercount == 0) {
					int percent;
					unsigned long l0, l1, l2;
					timercount++;
					user += nice;
					/* get current values for latter */
					l0 = (sys - prevOut) + (user - prevIn);
					l1 = prevTotal;
					/* save off new values */
					prevOut = sys;
					prevIn = user;
					prevTotal = sys + user + idle;
					/* now compute what we need to know */
					l2 = prevTotal - l1;
					if(l2 == 0) {
						percent = 0;
					} else {
						percent = ((200 * l0 / l2) + 1) / 2;
						if(percent > 100)
							percent = 100;
					}
					sprintf(ss, "%d%%", percent);
					Epplet_change_label(txtpercent, ss);
				} else {
					timercount++;
					if(timercount > 9)
						timercount = 0;
				}
			}
		}
		fclose(f);
		draw_flame();
		Epplet_paste_buf(buf, win, 0, 0);
		Epplet_timer(cb_timer, NULL, 0.1, "TIMER");
	}
	data = NULL;
}

/* called when you close the epplet */
static void cb_close(void *data)
{
	Epplet_unremember();
	Esync();
	data = NULL;
	exit(0);
}

/* called when you select a color from the epplet color menu */
static void cb_color(void *data)
{
	int *d;
   
	d = (int *)data;
	flame_col(d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7], d[8]);
	save_conf(d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7], d[8]);
	Epplet_gadget_hide(pop);
}

/* called when the config button is pressed */
static void cb_config(void *data)
{
	data = NULL;
	Epplet_gadget_show(pop);
}

/* called when the help button is pressed */
static void cb_help(void *data)
{
	data = NULL;
	Epplet_show_about("E-CPU2");
}

/* called when the epplet gets the focus */
static void cb_in(void *data, Window w)
{
	Epplet_gadget_show(b_close);
	Epplet_gadget_show(b_config);
	Epplet_gadget_show(b_help);
	return;
	data = NULL;
	w = (Window) 0;
}

/* called when the epplet loses the focus */
static void cb_out(void *data, Window w)
{
	Epplet_gadget_hide(b_close);
	Epplet_gadget_hide(b_config);
	Epplet_gadget_hide(b_help);
	return;
	data = NULL;
	w = (Window) 0;
}

/* save off the color of the flame in a config file */
static void save_conf(int d1, int d2, int d3, int d4, int d5,
					  int d6, int d7, int d8, int d9)
{
	char s[1024];
  
	sprintf(s, "%d %d %d", d1, d2, d3);
	Epplet_modify_config("color1", s);
	sprintf(s, "%d %d %d", d4, d5, d6);
	Epplet_modify_config("color2", s);
	sprintf(s, "%d %d %d", d7, d8, d9);
	Epplet_modify_config("color3", s);
	Epplet_save_config();
}

/* count the number of CPUs for division of flame area */
static int count_cpus(void)
{
	FILE *f;
	char s[256];
	
	f = fopen("/proc/stat", "r");
	if (f) {
		int count = 0;
		char ok = 1;
		
		while(ok) {
			if (!fgets(s, 255, f))
				ok = 0;
			else {
				if (strncmp(s, "cpu", 3))
					ok = 0;
				else
					count++;
			}
		}
		if (count > 1)
			count--;
		fclose (f);
		return count;
	}
	exit(1);
}


/* load the color of the flame from a config file */
static void load_conf(void)
{
	int d1, d2, d3, d4, d5, d6, d7, d8, d9;
	char *str;

	str = Epplet_query_config_def("color1", "30 90 90");
	sscanf(str, "%d %d %d", &d1, &d2, &d3);
	str = Epplet_query_config_def("color2", "50 255 255");
	sscanf(str, "%d %d %d", &d4, &d5, &d6);
	str = Epplet_query_config_def("color3", "255 255 255");
	sscanf(str, "%d %d %d", &d7, &d8, &d9);
	flame_col(d1, d2, d3,
			  d4, d5, d6,
			  d7, d8, d9);
}

/* where it all begins... */
int main(int argc, char **argv)
{
	Epplet_gadget p;
	char s[64];
	int prio;

	prio = getpriority(PRIO_PROCESS, getpid());
	setpriority(PRIO_PROCESS, getpid(), prio + 10);
	atexit(Epplet_cleanup);
  
	cpus = count_cpus();
	load_val = malloc(sizeof(int) * cpus);
	prev_val = malloc(sizeof(double) * cpus);
   
	Epplet_Init("E-CPU2", "0.2", "E CPU2 Epplet",
			3, 3, argc, argv, 0);
	Epplet_load_config();
	Epplet_timer(cb_timer, NULL, 0.1, "TIMER");
	Epplet_gadget_show(da = Epplet_create_drawingarea(2, 2, 
							(WIDTH + 4), (HEIGHT + 4)));
	win = Epplet_get_drawingarea_window(da);
	buf = Epplet_make_rgb_buf(WIDTH, HEIGHT);
	cur_cpu = (Epplet_get_instance() - 1) % cpus;
	sprintf(s, "CPU%d:", cur_cpu);
	Epplet_gadget_show(text = Epplet_create_label(2, 36, s, 1));
	Epplet_gadget_show(txtpercent = Epplet_create_label(-3, 36, "0%", 1));
	b_close = Epplet_create_button(NULL, NULL,
				3, 3, 0, 0, "CLOSE", 0, NULL,
				cb_close, NULL);
	b_help = Epplet_create_button(NULL, NULL,
				17, 3, 0, 0, "HELP", 0, NULL,
				cb_help, NULL);
	b_config = Epplet_create_button(NULL, NULL,
				31, 3, 0, 0, "CONFIGURE", 0, NULL,
				cb_config, NULL);
	p = Epplet_create_popup();
	Epplet_add_popup_entry(p, "Turquoise", NULL, cb_color, 
				(void *)(&(colors[0 * 9])));
	Epplet_add_popup_entry(p, "Fire     ", NULL, cb_color,
				(void *)(&(colors[1 * 9])));
	Epplet_add_popup_entry(p, "Copper"   , NULL, cb_color,
				(void *)(&(colors[2 * 9])));
	Epplet_add_popup_entry(p, "Violet"   , NULL, cb_color,
				(void *)(&(colors[3 * 9])));
	Epplet_add_popup_entry(p, "Night"    , NULL, cb_color,
				(void *)(&(colors[4 * 9])));
	Epplet_add_popup_entry(p, "Sunrise"  , NULL, cb_color,
				(void *)(&(colors[5 * 9])));
	Epplet_add_popup_entry(p, "Sunset"   , NULL, cb_color,
				(void *)(&(colors[6 * 9])));
	pop = Epplet_create_popupbutton("Colors", NULL, 5, 24, 36, 12, NULL, p);
	Epplet_register_focus_in_handler(cb_in, NULL);
	Epplet_register_focus_out_handler(cb_out, NULL);
	load_conf();
	Epplet_show();
	Epplet_Loop();
	return 0;
}

